@@ -1,4 +1,4 @@
-# Copyright 2004-2018 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2019 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -137,7 +137,11 @@ class ParameterInfo(object):
 
         if self.extrakw:
             rv[self.extrakw] = values
-        elif values and not ignore_errors:
+
+        elif values.get("_ignore_extra_kwargs", False):
+            pass
+
+        elif values and (not ignore_errors):
             raise Exception("Unknown keyword arguments: %s" % ( ", ".join(values.keys())))
 
         return rv
@@ -245,6 +249,16 @@ class PyExpr(unicode):
         return (unicode(self), self.filename, self.linenumber)  # E1101
 
 
+def probably_side_effect_free(expr):
+    """
+    Returns true if an expr probably does not have side effects, and should
+    be predicted. Basically, this just whitelists a set of characters that
+    doesn't allow for a function call.
+    """
+
+    return not ("(" in expr)
+
+
 class PyCode(object):
 
     __slots__ = [
@@ -574,6 +588,8 @@ class Say(Node):
         'interact',
         'attributes',
         'arguments',
+        'temporary_attributes',
+        'rollback',
         ]
 
     def diff_info(self):
@@ -584,9 +600,11 @@ class Say(Node):
         self.attributes = None
         self.interact = True
         self.arguments = None
+        self.temporary_attributes = None
+        self.rollback = "normal"
         return self
 
-    def __init__(self, loc, who, what, with_, interact=True, attributes=None, arguments=None):
+    def __init__(self, loc, who, what, with_, interact=True, attributes=None, arguments=None, temporary_attributes=None):
 
         super(Say, self).__init__(loc)
 
@@ -611,6 +629,9 @@ class Say(Node):
         # speaking, or None to disable this behavior.
         self.attributes = attributes
 
+        # Ditto for temporary attributes.
+        self.temporary_attributes = temporary_attributes
+
     def get_code(self, dialogue_filter=None):
         rv = [ ]
 
@@ -620,6 +641,10 @@ class Say(Node):
         if self.attributes is not None:
             rv.extend(self.attributes)
 
+        if self.temporary_attributes:
+            rv.append("@")
+            rv.extend(self.temporary_attributes)
+
         what = self.what
         if dialogue_filter is not None:
             what = dialogue_filter(what)
@@ -646,6 +671,7 @@ class Say(Node):
         try:
 
             renpy.game.context().say_attributes = self.attributes
+            renpy.game.context().temporary_attributes = self.temporary_attributes
 
             who = eval_who(self.who, self.who_fast)
 
@@ -668,6 +694,8 @@ class Say(Node):
                 args = tuple()
                 kwargs = dict()
 
+            kwargs.setdefault("interact", self.interact)
+
             if getattr(who, "record_say", True):
                 renpy.store._last_say_who = self.who
                 renpy.store._last_say_what = what
@@ -675,14 +703,16 @@ class Say(Node):
                 renpy.store._last_say_kwargs = kwargs
 
             say_menu_with(self.with_, renpy.game.interface.set_transition)
-            renpy.exports.say(who, what, interact=self.interact, *args, **kwargs)
+            renpy.exports.say(who, what, *args, **kwargs)
 
         finally:
             renpy.game.context().say_attributes = None
+            renpy.game.context().temporary_attributes = None
 
     def predict(self):
 
         old_attributes = renpy.game.context().say_attributes
+        old_temporary_attributes = renpy.game.context().temporary_attributes
 
         try:
 
@@ -703,6 +733,7 @@ class Say(Node):
 
         finally:
             renpy.game.context().say_attributes = old_attributes
+            renpy.game.context().temporary_attributes = old_temporary_attributes
 
         return [ self.next ]
 
@@ -767,7 +798,6 @@ class Label(Node):
     rollback = "force"
 
     translation_relevant = True
-
     __slots__ = [
         'name',
         'parameters',
@@ -1409,12 +1439,19 @@ class Call(Node):
             args, kwargs = self.arguments.evaluate()
             renpy.store._args = args
             renpy.store._kwargs = kwargs
+        else:
+            renpy.store._args = None
+            renpy.store._kwargs = None
 
     def predict(self):
 
         label = self.label
 
         if self.expression:
+
+            if not probably_side_effect_free(label):
+                return [ ]
+
             label = renpy.python.py_eval(label)
 
         return [ renpy.game.context().predict_call(label, self.next.name) ]
@@ -1487,20 +1524,28 @@ class Menu(Node):
         'set',
         'with_',
         'has_caption',
+        'arguments',
+        'item_arguments',
+        'rollback',
         ]
 
     def __new__(cls, *args, **kwargs):
         self = Node.__new__(cls)
         self.has_caption = False
+        self.arguments = None
+        self.item_arguments = None
+        self.rollback = "force"
         return self
 
-    def __init__(self, loc, items, set, with_, has_caption):  # @ReservedAssignment
+    def __init__(self, loc, items, set, with_, has_caption, arguments, item_arguments):  # @ReservedAssignment
         super(Menu, self).__init__(loc)
 
         self.items = items
         self.set = set
         self.with_ = with_
         self.has_caption = has_caption
+        self.arguments = arguments
+        self.item_arguments = item_arguments
 
     def diff_info(self):
         return (Menu,)
@@ -1538,28 +1583,47 @@ class Menu(Node):
         else:
             statement_name("menu")
 
+        if self.arguments is not None:
+            args, kwargs = self.arguments.evaluate()
+        else:
+            args = kwargs = None
+
         choices = [ ]
         narration = [ ]
+        item_arguments = [ ]
 
         for i, (label, condition, block) in enumerate(self.items):
 
             if renpy.config.say_menu_text_filter:
                 label = renpy.config.say_menu_text_filter(label)
 
+            has_item = False
+
             if block is None:
                 if renpy.config.narrator_menu and label:
                     narration.append(label)
                 else:
                     choices.append((label, condition, None))
+                    has_item = True
+
             else:
                 choices.append((label, condition, i))
+                has_item = True
+
                 next_node(block[0])
 
+            if has_item:
+                if self.item_arguments and (self.item_arguments[i] is not None):
+                    item_arguments.append(self.item_arguments[i].evaluate())
+                else:
+                    item_arguments.append((tuple(), dict()))
+
         if narration:
             renpy.exports.say(None, "\n".join(narration), interact=False)
 
         say_menu_with(self.with_, renpy.game.interface.set_transition)
-        choice = renpy.exports.menu(choices, self.set)
+
+        choice = renpy.exports.menu(choices, self.set, args, kwargs, item_arguments)
 
         if choice is not None:
             next_node(self.items[choice][2][0])
@@ -1635,10 +1699,16 @@ class Jump(Node):
 
     def predict(self):
 
+        label = self.target
+
         if self.expression:
-            return [ ]
-        else:
-            return [ renpy.game.script.lookup(self.target) ]
+
+            if not probably_side_effect_free(label):
+                return [ ]
+
+            label = renpy.python.py_eval(label)
+
+        return [ renpy.game.script.lookup(label) ]
 
     def scry(self):
         rv = Node.scry(self)
@@ -1784,6 +1854,9 @@ class UserStatement(Node):
         'block',
         'translatable',
         'code_block',
+        'translation_relevant',
+        'rollback',
+        'subparses',
         ]
 
     def __new__(cls, *args, **kwargs):
@@ -1791,20 +1864,22 @@ class UserStatement(Node):
         self.block = [ ]
         self.code_block = None
         self.translatable = False
+        self.translation_relevant = False
+        self.rollback = "normal"
+        self.subparses = [ ]
         return self
 
-    def __init__(self, loc, line, block):
+    def __init__(self, loc, line, block, parsed):
 
         super(UserStatement, self).__init__(loc)
+        self.code_block = None
+        self.parsed = parsed
         self.line = line
         self.block = block
-        self.code_block = None
-        self.parsed = None
+        self.subparses = [ ]
 
         self.name = self.call("label")
-
-        # Do not store the parse.
-        self.parsed = None
+        self.rollback = renpy.statements.get("rollback", self.parsed) or "normal"
 
     def __repr__(self):
         return "<UserStatement {!r}>".format(self.line)
@@ -1812,11 +1887,13 @@ class UserStatement(Node):
     def get_children(self, f):
         f(self)
 
-        if not self.code_block:
-            return
+        if self.code_block is not None:
+            for i in self.code_block:
+                i.get_children(f)
 
-        for i in self.code_block:
-            i.get_children(f)
+        for i in self.subparses:
+            for j in i.block:
+                j.get_children(f)
 
     def chain(self, next):  # @ReservedAssignment
         self.next = next
@@ -1824,22 +1901,33 @@ class UserStatement(Node):
         if self.code_block is not None:
             chain_block(self.code_block, next)
 
+        for i in self.subparses:
+            chain_block(i.block, next)
+
     def replace_next(self, old, new):
         Node.replace_next(self, old, new)
 
         if (self.code_block) and (self.code_block[0] is old):
             self.code_block.insert(0, new)
 
+        for i in self.subparses:
+            if i.block[0] is old:
+                self.code_block.insert(0, new)
+
     def restructure(self, callback):
         if self.code_block:
             callback(self.code_block)
 
+        for i in self.subparses:
+            callback(i.block)
+
     def diff_info(self):
         return (UserStatement, self.line)
 
     def call(self, method, *args, **kwargs):
 
         parsed = self.parsed
+
         if parsed is None:
             parsed = renpy.statements.parse(self, self.line, self.block)
             self.parsed = parsed
@@ -1865,10 +1953,25 @@ class UserStatement(Node):
             for i in predictions:
                 renpy.easy.predict(i)
 
-        return [ self.get_next() ]
+        if self.parsed and renpy.statements.get("predict_all", self.parsed):
+            return [ i.block[0] for i in self.subparses ] + [ self.next ]
+
+        if self.next:
+            next_label = self.next.name
+        else:
+            next_label = None
+
+        next_list = self.call("predict_next", next_label)
+
+        if next_list is not None:
+            nexts = [ renpy.game.script.lookup_or_none(i) for i in next_list if i is not None ]
+            return [ i for i in nexts if i is not None ]
+
+        return [ self.next ]
 
     def get_name(self):
         parsed = self.parsed
+
         if parsed is None:
             parsed = renpy.statements.parse(self, self.line, self.block)
             self.parsed = parsed
@@ -1904,6 +2007,32 @@ class UserStatement(Node):
         return False
 
 
+class PostUserStatement(Node):
+
+    __slots__ = [
+        'parent',
+        ]
+
+    def __init__(self, loc, parent):
+
+        super(PostUserStatement, self).__init__(loc)
+        self.parent = parent
+
+        self.name = self.parent.call("post_label")
+
+    def __repr__(self):
+        return "<PostUserStatement {!r}>".format(self.parent.line)
+
+    def diff_info(self):
+        return (PostUserStatement, self.parent.line)
+
+    def execute(self):
+        next_node(self.next)
+        statement_name("post " + self.parent.get_name())
+
+        self.parent.call("post_execute")
+
+
 def create_store(name):
     if name not in renpy.config.special_namespaces:
         renpy.python.create_store(name)
@@ -1932,7 +2061,7 @@ def get_namespace(store):
 
 # Config variables that are set twice - once when the rpy is first loaded,
 # and then again at init time.
-EARLY_CONFIG = { "save_directory" }
+EARLY_CONFIG = { "save_directory", "allow_duplicate_labels", "keyword_after_python" }
 
 define_statements = [ ]
 
