@@ -1,4 +1,4 @@
-# Copyright 2004-2015 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2019 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -19,18 +19,22 @@
 # OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 # WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
+from __future__ import print_function
+
 import renpy.display
 import renpy.pyanalysis
 
 import random
 
+
 def compiling(loc):
-    file, number = loc #@ReservedAssignment
+    file, number = loc  # @ReservedAssignment
 
     renpy.game.exception_info = "Compiling ATL code at %s:%d" % (file, number)
 
+
 def executing(loc):
-    file, number = loc #@ReservedAssignment
+    file, number = loc  # @ReservedAssignment
 
     renpy.game.exception_info = "Executing ATL code at %s:%d" % (file, number)
 
@@ -38,6 +42,7 @@ def executing(loc):
 # A map from the name of a time warp function to the function itself.
 warpers = { }
 
+
 def atl_warper(f):
     name = f.func_name
     warpers[name] = f
@@ -45,6 +50,8 @@ def atl_warper(f):
 
 # The pause warper is used internally when no other warper is
 # specified.
+
+
 @atl_warper
 def pause(t):
     if t < 1.0:
@@ -52,60 +59,80 @@ def pause(t):
     else:
         return 1.0
 
+
+@atl_warper
+def instant(t):
+    return 1.0
+
+
 position = renpy.object.Sentinel("position")
 
+
 def any_object(x):
     return x
 
+
 def bool_or_none(x):
     if x is None:
         return x
     return bool(x)
 
+
+def float_or_none(x):
+    if x is None:
+        return x
+    return float(x)
+
+
 # A dictionary giving property names and the corresponding default
 # values.
 PROPERTIES = {
-        "pos" : (position, position),
-        "xpos" : position,
-        "ypos" : position,
-        "anchor" : (position, position),
-        "xanchor" : position,
-        "yanchor" : position,
-        "xaround" : position,
-        "yaround" : position,
-        "xanchoraround" : float,
-        "yanchoraround" : float,
-        "align" : (float, float),
-        "xalign" : float,
-        "yalign" : float,
-        "rotate" : float,
-        "rotate_pad" : bool,
-        "transform_anchor" : bool,
-        "xzoom" : float,
-        "yzoom" : float,
-        "zoom" : float,
-        "nearest" : bool_or_none,
-        "alpha" : float,
-        "additive" : float,
-        "around" : (position, position),
-        "alignaround" : (float, float),
-        "angle" : float,
-        "radius" : float,
-        "crop" : (float, float, float, float),
-        "crop_relative" : bool,
-        "size" : (int, int),
-        "corner1" : (float, float),
-        "corner2" : (float, float),
-        "subpixel" : bool,
-        "delay" : float,
-        "xoffset" : float,
-        "yoffset" : float,
-        "offset" : (int, int),
-        "xcenter" : position,
-        "ycenter" : position,
-        "debug" : any_object,
-        "events" : bool,
-        }
+    "pos" : (position, position),
+    "xpos" : position,
+    "ypos" : position,
+    "anchor" : (position, position),
+    "xanchor" : position,
+    "yanchor" : position,
+    "xaround" : position,
+    "yaround" : position,
+    "xanchoraround" : float,
+    "yanchoraround" : float,
+    "align" : (float, float),
+    "xalign" : float,
+    "yalign" : float,
+    "rotate" : float,
+    "rotate_pad" : bool,
+    "transform_anchor" : bool,
+    "xzoom" : float,
+    "yzoom" : float,
+    "zoom" : float,
+    "nearest" : bool_or_none,
+    "alpha" : float,
+    "additive" : float,
+    "around" : (position, position),
+    "alignaround" : (float, float),
+    "angle" : float,
+    "radius" : float,
+    "crop" : (float, float, float, float),
+    "crop_relative" : bool,
+    "size" : (int, int),
+    "maxsize" : (int, int),
+    "corner1" : (float, float),
+    "corner2" : (float, float),
+    "subpixel" : bool,
+    "delay" : float,
+    "xoffset" : float,
+    "yoffset" : float,
+    "offset" : (int, int),
+    "xcenter" : position,
+    "ycenter" : position,
+    "debug" : any_object,
+    "events" : bool,
+    "xpan" : float_or_none,
+    "ypan" : float_or_none,
+    "xtile" : int,
+    "ytile" : int,
+    }
 
 
 def correct_type(v, b, ty):
@@ -122,14 +149,11 @@ def correct_type(v, b, ty):
         return ty(v)
 
 
-def interpolate(t, a, b, type): #@ReservedAssignment
+def interpolate(t, a, b, type):  # @ReservedAssignment
     """
     Linearly interpolate the arguments.
     """
 
-    if t >= 1.0:
-        return b
-
     # Recurse into tuples.
     if isinstance(b, tuple):
         if a is None:
@@ -138,7 +162,7 @@ def interpolate(t, a, b, type): #@ReservedAssignment
         return tuple(interpolate(t, i, j, ty) for i, j, ty in zip(a, b, type))
 
     # Deal with booleans, nones, etc.
-    elif b is None or isinstance(b, bool):
+    elif b is None or isinstance(b, (bool, basestring)):
         if t >= 1.0:
             return b
         else:
@@ -153,6 +177,8 @@ def interpolate(t, a, b, type): #@ReservedAssignment
 
 # Interpolate the value of a spline. This code is based on Aenakume's code,
 # from 00splines.rpy.
+
+
 def interpolate_spline(t, spline):
 
     if isinstance(spline[-1], tuple):
@@ -187,9 +213,11 @@ def interpolate_spline(t, spline):
 
     return correct_type(rv, spline[-1], position)
 
+
 # A list of atl transforms that may need to be compile.
 compile_queue = [ ]
 
+
 def compile_all():
     """
     Called after the init phase is finished and transforms are compiled,
@@ -209,12 +237,13 @@ def compile_all():
 # scopes that are used to evaluate the various expressions in the statement,
 # and has a method to do the evaluation and return a result.
 class Context(object):
+
     def __init__(self, context):
         self.context = context
 
-    def eval(self, expr): #@ReservedAssignment
+    def eval(self, expr):  # @ReservedAssignment
         expr = renpy.python.escape_unicode(expr)
-        return eval(expr, renpy.store.__dict__, self.context) #@UndefinedVariable
+        return eval(expr, renpy.store.__dict__, self.context)  # @UndefinedVariable
 
     def __eq__(self, other):
         if not isinstance(other, Context):
@@ -222,9 +251,14 @@ class Context(object):
 
         return self.context == other.context
 
+    def __ne__(self, other):
+        return not (self == other)
+
 # This is intended to be subclassed by ATLTransform. It takes care of
 # managing ATL execution, which allows ATLTransform itself to not care
 # much about the contents of this file.
+
+
 class ATLTransformBase(renpy.object.Object):
 
     # Compatibility with older saves.
@@ -293,6 +327,15 @@ class ATLTransformBase(renpy.object.Object):
         if renpy.game.context().init_phase:
             compile_queue.append(self)
 
+    def _handles_event(self, event):
+        if (self.block is not None) and (self.block._handles_event(event)):
+            return True
+
+        if self.child is None:
+            return False
+
+        return self.child._handles_event(event)
+
     def get_block(self):
         """
         Returns the compiled block to use.
@@ -311,14 +354,13 @@ class ATLTransformBase(renpy.object.Object):
         requires that t.atl is self.atl.
         """
 
-        if self is t:
-            return
-
         super(ATLTransformBase, self).take_execution_state(t)
 
         self.atl_st_offset = None
 
-        if not isinstance(t, ATLTransformBase):
+        if self is t:
+            return
+        elif not isinstance(t, ATLTransformBase):
             return
         elif t.atl is not self.atl:
             return
@@ -346,11 +388,18 @@ class ATLTransformBase(renpy.object.Object):
         self.atl_st_offset = t.atl_st_offset
 
         if self.child is renpy.display.motion.null:
-            self.child = t.child
+
+            if t.child and t.child._duplicatable:
+                self.child = t.child._duplicate(None)
+            else:
+                self.child = t.child
+
             self.raw_child = t.raw_child
 
     def __call__(self, *args, **kwargs):
 
+        _args = kwargs.pop("_args", None)
+
         context = self.context.context.copy()
 
         for k, v in self.parameters.parameters:
@@ -394,9 +443,6 @@ class ATLTransformBase(renpy.object.Object):
         if child is None:
             child = self.child
 
-        if child is None:
-            child = renpy.display.motion.get_null()
-
         # Create a new ATL Transform.
         parameters = renpy.ast.ParameterInfo({ }, positional, None, None)
 
@@ -405,15 +451,16 @@ class ATLTransformBase(renpy.object.Object):
             child=child,
             style=self.style_arg,
             context=context,
-            parameters=parameters)
+            parameters=parameters,
+            _args=_args,
+            )
 
         rv.parent_transform = self
         rv.take_state(self)
 
         return rv
 
-
-    def compile(self): #@ReservedAssignment
+    def compile(self):  # @ReservedAssignment
         """
         Compiles the ATL code into a block. As necessary, updates the
         properties.
@@ -441,12 +488,13 @@ class ATLTransformBase(renpy.object.Object):
 
         block = self.atl.compile(self.context)
 
-        if len(block.statements) == 1 and isinstance(block.statements[0], Interpolation):
-
-            interp = block.statements[0]
-
-            if interp.duration == 0 and interp.properties:
-                self.properties = interp.properties[:]
+        if all(
+            isinstance(statement, Interpolation) and statement.duration == 0
+            for statement in block.statements
+        ):
+            self.properties = []
+            for interp in block.statements:
+                self.properties.extend(interp.properties)
 
         if not constant and renpy.display.predict.predicting:
             self.predict_block = block
@@ -463,7 +511,6 @@ class ATLTransformBase(renpy.object.Object):
 
         return block
 
-
     def execute(self, trans, st, at):
 
         if self.done:
@@ -473,11 +520,7 @@ class ATLTransformBase(renpy.object.Object):
         if block is None:
             block = self.compile()
 
-        # Propagate transform_events from children.
-        if self.child:
-            if self.child.transform_event != self.last_child_transform_event:
-                self.last_child_transform_event = self.child.transform_event
-                self.transform_event = self.child.transform_event
+        events = [ ]
 
         # Hide request.
         if trans.hide_request:
@@ -487,11 +530,26 @@ class ATLTransformBase(renpy.object.Object):
             self.transform_event = "replaced"
 
         # Notice transform events.
+        if renpy.config.atl_multiple_events:
+            if self.transform_event != self.last_transform_event:
+                events.append(self.transform_event)
+                self.last_transform_event = self.transform_event
+
+        # Propagate transform_events from children.
+        if (self.child is not None) and self.child.transform_event != self.last_child_transform_event:
+            self.last_child_transform_event = self.child.transform_event
+
+            if self.child.transform_event is not None:
+                self.transform_event = self.child.transform_event
+
+        # Notice transform events, again.
         if self.transform_event != self.last_transform_event:
-            event = self.transform_event
+            events.append(self.transform_event)
             self.last_transform_event = self.transform_event
-        else:
-            event = None
+
+        if self.transform_event in renpy.config.repeat_transform_events:
+            self.transform_event = None
+            self.last_transform_event = None
 
         old_exception_info = renpy.game.exception_info
 
@@ -503,7 +561,7 @@ class ATLTransformBase(renpy.object.Object):
         else:
             timebase = st - self.atl_st_offset
 
-        action, arg, pause = block.execute(trans, timebase, self.atl_state, event)
+        action, arg, pause = block.execute(trans, timebase, self.atl_state, events)
 
         renpy.game.exception_info = old_exception_info
 
@@ -525,11 +583,14 @@ class ATLTransformBase(renpy.object.Object):
 
         return self.children + block.visit()
 
+
 # This is used in mark_constant to analyze expressions for constness.
 is_constant_expr = renpy.pyanalysis.Analysis().is_constant_expr
 GLOBAL_CONST = renpy.pyanalysis.GLOBAL_CONST
 
 # The base class for raw ATL statements.
+
+
 class RawStatement(object):
 
     constant = None
@@ -540,7 +601,7 @@ class RawStatement(object):
 
     # Compiles this RawStatement into a Statement, by using ctx to
     # evaluate expressions as necessary.
-    def compile(self, ctx): #@ReservedAssignment
+    def compile(self, ctx):  # @ReservedAssignment
         raise Exception("Compile not implemented.")
 
     # Predicts the images used by this statement.
@@ -556,6 +617,8 @@ class RawStatement(object):
         self.constant = 0
 
 # The base class for compiled ATL Statements.
+
+
 class Statement(renpy.object.Object):
 
     def __init__(self, loc):
@@ -587,14 +650,20 @@ class Statement(renpy.object.Object):
     #
     # Pause is the amount of time until execute should be called again,
     # or None if there's no need to call execute ever again.
-    def execute(self, trans, st, state, event):
+    def execute(self, trans, st, state, events):
         raise Exception("Not implemented.")
 
     # Return a list of displayable children.
     def visit(self):
         return [ ]
 
+    # Does this respond to an event?
+    def _handles_event(self, event):
+        return False
+
 # This represents a Raw ATL block.
+
+
 class RawBlock(RawStatement):
 
     # Should we use the animation timebase or the showing timebase?
@@ -609,7 +678,7 @@ class RawBlock(RawStatement):
 
         self.animation = animation
 
-    def compile(self, ctx): #@ReservedAssignment
+    def compile(self, ctx):  # @ReservedAssignment
         compiling(self.loc)
 
         statements = [ i.compile(ctx) for i in self.statements ]
@@ -633,6 +702,7 @@ class RawBlock(RawStatement):
 
 # A compiled ATL block.
 class Block(Statement):
+
     def __init__(self, loc, statements):
 
         super(Block, self).__init__(loc)
@@ -649,7 +719,15 @@ class Block(Statement):
 
         self.times.sort()
 
-    def execute(self, trans, st, state, event):
+    def _handles_event(self, event):
+
+        for i in self.statements:
+            if i._handles_event(event):
+                return True
+
+        return False
+
+    def execute(self, trans, st, state, events):
 
         executing(self.loc)
 
@@ -688,10 +766,9 @@ class Block(Statement):
                 if index >= len(self.statements):
                     return "next", target - start, None
 
-
                 # Find the statement and try to run it.
                 stmt = self.statements[index]
-                action, arg, pause = stmt.execute(trans, target - start, child_state, event)
+                action, arg, pause = stmt.execute(trans, target - start, child_state, events)
 
                 # On continue, persist our state.
                 if action == "continue":
@@ -764,6 +841,8 @@ class Block(Statement):
 #
 # We won't decide which it is until runtime, as we need the
 # values of the variables here.
+
+
 class RawMultipurpose(RawStatement):
 
     warp_function = None
@@ -800,7 +879,7 @@ class RawMultipurpose(RawStatement):
     def add_spline(self, name, exprs):
         self.splines.append((name, exprs))
 
-    def compile(self, ctx): #@ReservedAssignment
+    def compile(self, ctx):  # @ReservedAssignment
 
         compiling(self.loc)
 
@@ -811,7 +890,7 @@ class RawMultipurpose(RawStatement):
             self.warp_function is None and
             not self.properties and
             not self.splines and
-            len(self.expressions) == 1):
+                len(self.expressions) == 1):
 
             expr, withexpr = self.expressions[0]
 
@@ -839,7 +918,7 @@ class RawMultipurpose(RawStatement):
         if self.warp_function:
             warper = ctx.eval(self.warp_function)
         else:
-            warper = self.warper or "pause"
+            warper = self.warper or "instant"
 
             if warper not in warpers:
                 raise Exception("ATL Warper %s is unknown at runtime." % warper)
@@ -877,7 +956,6 @@ class RawMultipurpose(RawStatement):
             if value.properties is None:
                 raise Exception("ATL transform %r is too complicated to be included in interpolation." % expr)
 
-
             properties.extend(value.properties)
 
         duration = ctx.eval(self.duration)
@@ -905,7 +983,6 @@ class RawMultipurpose(RawStatement):
 
         self.constant = constant
 
-
     def predict(self, ctx):
 
         for i, _j in self.expressions:
@@ -925,6 +1002,8 @@ class RawMultipurpose(RawStatement):
                 continue
 
 # This lets us have an ATL transform as our child.
+
+
 class RawContainsExpr(RawStatement):
 
     def __init__(self, loc, expr):
@@ -933,7 +1012,7 @@ class RawContainsExpr(RawStatement):
 
         self.expression = expr
 
-    def compile(self, ctx): #@ReservedAssignment
+    def compile(self, ctx):  # @ReservedAssignment
         compiling(self.loc)
         child = ctx.eval(self.expression)
         return Child(self.loc, child, None)
@@ -951,7 +1030,7 @@ class RawChild(RawStatement):
 
         self.children = [ child ]
 
-    def compile(self, ctx): #@ReservedAssignment
+    def compile(self, ctx):  # @ReservedAssignment
 
         children = [ ]
 
@@ -965,7 +1044,6 @@ class RawChild(RawStatement):
 
         return Child(self.loc, box, None)
 
-
     def mark_constant(self):
 
         constant = GLOBAL_CONST
@@ -987,20 +1065,26 @@ class Child(Statement):
         self.child = child
         self.transition = transition
 
-    def execute(self, trans, st, state, event):
+    def execute(self, trans, st, state, events):
 
         executing(self.loc)
 
         old_child = trans.raw_child
-        child = self.child.parameterize('displayable', [ ])
+
+        child = self.child
+
+        if child._duplicatable:
+            child = self.child._duplicate(trans._args)
+            child._unique()
 
         if (old_child is not None) and (old_child is not renpy.display.motion.null) and (self.transition is not None):
             child = self.transition(old_widget=old_child,
                                     new_widget=child)
+            child._unique()
         else:
             child = child
 
-        trans.set_child(child)
+        trans.set_child(child, duplicate=False)
         trans.raw_child = self.child
 
         return "next", st, None
@@ -1027,17 +1111,29 @@ class Interpolation(Statement):
         # The number of complete circles we make.
         self.circles = circles
 
-    def execute(self, trans, st, state, event):
+    def execute(self, trans, st, state, events):
 
         executing(self.loc)
 
         warper = warpers.get(self.warper, self.warper)
 
+        if (self.warper != "instant") and (state is None) and (
+                (trans.atl_state is not None) or (trans.st == 0)
+                ):
+            first = True
+        else:
+            first = False
+
         if self.duration:
             complete = min(1.0, st / self.duration)
         else:
             complete = 1.0
 
+        if complete < 0.0:
+            complete = 0.0
+        elif complete > 1.0:
+            complete = 1.0
+
         complete = warper(complete)
 
         if state is None:
@@ -1047,9 +1143,15 @@ class Interpolation(Statement):
             newts = renpy.display.motion.TransformState()
             newts.take_state(trans.state)
 
+            has_angle = False
+
             for k, v in self.properties:
                 setattr(newts, k, v)
 
+                if k == "angle":
+                    newts.last_angle = v
+                    has_angle = True
+
             # Now, the things we change linearly are in the difference
             # between the new and old states.
             linear = trans.state.diff(newts)
@@ -1057,8 +1159,10 @@ class Interpolation(Statement):
             revolution = None
             splines = [ ]
 
-            # Clockwise revolution.
-            if self.revolution is not None:
+            revdir = self.revolution
+            circles = self.circles
+
+            if (revdir or (has_angle and renpy.config.automatic_polar_motion)) and (newts.xaround is not None):
 
                 # Remove various irrelevant motions.
                 for i in [ 'xpos', 'ypos',
@@ -1069,7 +1173,7 @@ class Interpolation(Statement):
 
                     linear.pop(i, None)
 
-                if newts.xaround is not None:
+                if revdir is not None:
 
                     # Ensure we rotate around the new point.
                     trans.state.xaround = newts.xaround
@@ -1086,21 +1190,26 @@ class Interpolation(Statement):
                     # Make sure the revolution is in the appropriate direction,
                     # and contains an appropriate number of circles.
 
-                    if self.revolution == "clockwise":
+                    if revdir == "clockwise":
                         if endangle < startangle:
                             startangle -= 360
 
-                        startangle -= self.circles * 360
+                        startangle -= circles * 360
 
-                    elif self.revolution == "counterclockwise":
+                    elif revdir == "counterclockwise":
                         if endangle > startangle:
                             startangle += 360
 
-                        startangle += self.circles * 360
+                        startangle += circles * 360
 
                     # Store the revolution.
                     revolution = (startangle, endangle, startradius, endradius)
 
+                else:
+
+                    last_angle = trans.state.last_angle or trans.state.angle
+                    revolution = (last_angle, newts.last_angle, trans.state.radius, newts.radius)
+
             # Figure out the splines.
             for name, values in self.splines:
                 splines.append((name, [ getattr(trans.state, name) ] + values))
@@ -1125,20 +1234,23 @@ class Interpolation(Statement):
         # Handle the revolution.
         if revolution is not None:
             startangle, endangle, startradius, endradius = revolution
-            trans.state.angle = interpolate(complete, startangle, endangle, float)
-            trans.state.radius = interpolate(complete, startradius, endradius, float)
 
+            angle = interpolate(complete, startangle, endangle, float)
+            trans.state.last_angle = angle
+            trans.state.angle = angle
+
+            trans.state.radius = interpolate(complete, startradius, endradius, float)
 
         # Handle any splines we might have.
         for name, values in splines:
             value = interpolate_spline(complete, values)
             setattr(trans.state, name, value)
 
-        if st >= self.duration:
+        if ((not first) or (not renpy.config.atl_one_frame)) and (st >= self.duration):
             return "next", st - self.duration, None
         else:
             if not self.properties and not self.revolution and not self.splines:
-                return "continue", state, self.duration - st
+                return "continue", state, max(0, self.duration - st)
             else:
                 return "continue", state, 0
 
@@ -1152,7 +1264,7 @@ class RawRepeat(RawStatement):
 
         self.repeats = repeats
 
-    def compile(self, ctx): #@ReservedAssignment
+    def compile(self, ctx):  # @ReservedAssignment
 
         compiling(self.loc)
 
@@ -1166,6 +1278,7 @@ class RawRepeat(RawStatement):
     def mark_constant(self):
         self.constant = is_constant_expr(self.repeats)
 
+
 class Repeat(Statement):
 
     def __init__(self, loc, repeats):
@@ -1174,7 +1287,7 @@ class Repeat(Statement):
 
         self.repeats = repeats
 
-    def execute(self, trans, st, state, event):
+    def execute(self, trans, st, state, events):
         return "repeat", (self.repeats, st), 0
 
 
@@ -1187,7 +1300,7 @@ class RawParallel(RawStatement):
         super(RawParallel, self).__init__(loc)
         self.blocks = [ block ]
 
-    def compile(self, ctx): #@ReservedAssignment
+    def compile(self, ctx):  # @ReservedAssignment
         return Parallel(self.loc, [i.compile(ctx) for i in self.blocks])
 
     def predict(self, ctx):
@@ -1203,13 +1316,22 @@ class RawParallel(RawStatement):
 
         self.constant = constant
 
+
 class Parallel(Statement):
 
     def __init__(self, loc, blocks):
         super(Parallel, self).__init__(loc)
         self.blocks = blocks
 
-    def execute(self, trans, st, state, event):
+    def _handles_event(self, event):
+
+        for i in self.blocks:
+            if i._handles_event(event):
+                return True
+
+        return False
+
+    def execute(self, trans, st, state, events):
 
         executing(self.loc)
 
@@ -1227,7 +1349,7 @@ class Parallel(Statement):
 
         for i, istate in state:
 
-            action, arg, pause = i.execute(trans, st, istate, event)
+            action, arg, pause = i.execute(trans, st, istate, events)
 
             if pause is not None:
                 pauses.append(pause)
@@ -1257,7 +1379,7 @@ class RawChoice(RawStatement):
 
         self.choices = [ (chance, block) ]
 
-    def compile(self, ctx): #@ReservedAssignment
+    def compile(self, ctx):  # @ReservedAssignment
         compiling(self.loc)
         return Choice(self.loc, [ (ctx.eval(chance), block.compile(ctx)) for chance, block in self.choices])
 
@@ -1274,6 +1396,7 @@ class RawChoice(RawStatement):
 
         self.constant = constant
 
+
 class Choice(Statement):
 
     def __init__(self, loc, choices):
@@ -1282,7 +1405,15 @@ class Choice(Statement):
 
         self.choices = choices
 
-    def execute(self, trans, st, state, event):
+    def _handles_event(self, event):
+
+        for i in self.choices:
+            if i[1]._handles_event(event):
+                return True
+
+        return False
+
+    def execute(self, trans, st, state, events):
 
         executing(self.loc)
 
@@ -1304,7 +1435,7 @@ class Choice(Statement):
         else:
             choice, cstate = state
 
-        action, arg, pause = choice.execute(trans, st, cstate, event)
+        action, arg, pause = choice.execute(trans, st, cstate, events)
 
         if action == "continue":
             return "continue", (choice, arg), pause
@@ -1324,13 +1455,14 @@ class RawTime(RawStatement):
         super(RawTime, self).__init__(loc)
         self.time = time
 
-    def compile(self, ctx): #@ReservedAssignment
+    def compile(self, ctx):  # @ReservedAssignment
         compiling(self.loc)
         return Time(self.loc, ctx.eval(self.time))
 
     def mark_constant(self):
         self.constant = is_constant_expr(self.time)
 
+
 class Time(Statement):
 
     def __init__(self, loc, time):
@@ -1338,7 +1470,7 @@ class Time(Statement):
 
         self.time = time
 
-    def execute(self, trans, st, state, event):
+    def execute(self, trans, st, state, events):
         return "continue", None, None
 
 
@@ -1354,8 +1486,7 @@ class RawOn(RawStatement):
         for i in names:
             self.handlers[i] = block
 
-    def compile(self, ctx): #@ReservedAssignment
-
+    def compile(self, ctx):  # @ReservedAssignment
         compiling(self.loc)
 
         handlers = { }
@@ -1386,7 +1517,13 @@ class On(Statement):
 
         self.handlers = handlers
 
-    def execute(self, trans, st, state, event):
+    def _handles_event(self, event):
+        if event in self.handlers:
+            return True
+        else:
+            return False
+
+    def execute(self, trans, st, state, events):
 
         executing(self.loc)
 
@@ -1398,15 +1535,17 @@ class On(Statement):
 
         # If we have an external event, and we have a handler for it,
         # handle it.
-        if event in self.handlers:
+        for event in events:
 
-            # Do not allow people to abort the hide or replaced event.
-            lock_event = (name == "hide" and trans.hide_request) or (name == "replaced" and trans.replaced_request)
+            if event in self.handlers:
 
-            if not lock_event:
-                name = event
-                start = st
-                cstate = None
+                # Do not allow people to abort the hide or replaced event.
+                lock_event = (name == "hide" and trans.hide_request) or (name == "replaced" and trans.replaced_request)
+
+                if not lock_event:
+                    name = event
+                    start = st
+                    cstate = None
 
         while True:
 
@@ -1414,7 +1553,7 @@ class On(Statement):
             if name not in self.handlers:
                 return "continue", (name, start, cstate), None
 
-            action, arg, pause = self.handlers[name].execute(trans, st - start, cstate, event)
+            action, arg, pause = self.handlers[name].execute(trans, st - start, cstate, events)
 
             # If we get a continue, save our state.
             if action == "continue":
@@ -1466,12 +1605,13 @@ class RawEvent(RawStatement):
 
         self.name = name
 
-    def compile(self, ctx): #@ReservedAssignment
+    def compile(self, ctx):  # @ReservedAssignment
         return Event(self.loc, self.name)
 
     def mark_constant(self):
         self.constant = GLOBAL_CONST
 
+
 class Event(Statement):
 
     def __init__(self, loc, name):
@@ -1479,7 +1619,7 @@ class Event(Statement):
 
         self.name = name
 
-    def execute(self, trans, st, state, event):
+    def execute(self, trans, st, state, events):
         return "event", (self.name, st), None
 
 
@@ -1490,13 +1630,14 @@ class RawFunction(RawStatement):
 
         self.expr = expr
 
-    def compile(self, ctx): #@ReservedAssignment
+    def compile(self, ctx):  # @ReservedAssignment
         compiling(self.loc)
         return Function(self.loc, ctx.eval(self.expr))
 
     def mark_constant(self):
         self.constant = is_constant_expr(self.expr)
 
+
 class Function(Statement):
 
     def __init__(self, loc, function):
@@ -1504,7 +1645,10 @@ class Function(Statement):
 
         self.function = function
 
-    def execute(self, trans, st, state, event):
+    def _handles_event(self, event):
+        return True
+
+    def execute(self, trans, st, state, events):
         fr = self.function(trans, st, trans.at)
 
         if fr is not None:
@@ -1643,7 +1787,6 @@ def parse_atl(l):
             cp = l.checkpoint()
             warper = l.name()
 
-
             if warper in warpers:
                 duration = l.require(l.simple_expression)
                 warp_function = None
@@ -1737,14 +1880,12 @@ def parse_atl(l):
 
             statements.append(rm)
 
-
         if l.eol():
             l.advance()
             continue
 
         l.require(",", "comma or end of line")
 
-
     # Merge together statements that need to be merged together.
 
     merged = [ ]
