@@ -263,6 +263,11 @@ class Script(object):
         initcode = [ ]
 
         for fn, dir in script_files:  # @ReservedAssignment
+            # Mitigate "busy script" warning from the browser
+            if renpy.emscripten:
+                import emscripten
+                emscripten.sleep(0)
+
             self.load_appropriate_file(".rpyc", ".rpy", dir, fn, initcode)
 
         # Make the sort stable.
@@ -873,6 +878,9 @@ class Script(object):
         raises a ScriptError.
         """
 
+        if isinstance(label, renpy.parser.SubParse):
+            label = label.block[0].name
+
         label = renpy.config.label_overrides.get(label, label)
         original = label
 
@@ -892,10 +900,30 @@ class Script(object):
         Returns true if the label exists, or false otherwise.
         """
 
+        if isinstance(label, renpy.parser.SubParse):
+
+            if not label.block:
+                return False
+
+            label = label.block[0].name
+
         label = renpy.config.label_overrides.get(label, label)
 
         return label in self.namemap
 
+    def lookup_or_none(self, label):
+        """
+        Looks up the label if it exists, or returns None if it does not.
+        """
+
+        if label is None:
+            return None
+
+        if not self.has_label(label):
+            return None
+
+        return self.lookup(label)
+
     def analyze(self):
         """
         Analyzes all statements that need analysis.
